'Test of energy conservation in long VelocityVerlet run.'

from asap3 import EMT, MakeParallelAtoms, VelocityVerlet, Trajectory, MDLogger, units
from ase.lattice.cubic import FaceCenteredCubic
import numpy as np
from asap3.mpi import world
from asap3.md.velocitydistribution import MaxwellBoltzmannDistribution
from asap3.test.pytest_markers import ReportTest

import pytest

writetraj = True

@pytest.mark.slow
@pytest.mark.parametrize('boundary', [(1,1,1), (0,1,1)])
def test_longverlet(boundary, cpulayout, in_tmp_dir):
        ismaster = world.rank == 0
        isparallel = world.size != 1
        print ("CPU Layout: %s.  Periodic boundary conditions: %s."
               % (str(cpulayout), str(boundary)))
        if ismaster:
            size = np.array((11, 8, 7))
            if cpulayout:
                size *= cpulayout
            atoms = FaceCenteredCubic(size=size, symbol="Cu",
                                      pbc=boundary, latticeconstant=3.61*1.04)
        else:
            atoms = None
        if isparallel:
            atoms = MakeParallelAtoms(atoms, cpulayout)
        natoms = atoms.get_global_number_of_atoms()
        atoms.calc = EMT()
        MaxwellBoltzmannDistribution(atoms, temperature_K=3000)
        if ismaster:
            print("Initializing")

        dyn = VelocityVerlet(atoms, logfile="-", timestep=3*units.fs, loginterval=1)
        dyn.run(50)
        e_start = (atoms.get_potential_energy() 
                   + atoms.get_kinetic_energy())/natoms
        if ismaster:
            print("Running")
        dyn = VelocityVerlet(atoms, timestep=5*units.fs)
        logger = MDLogger(dyn, atoms, '-', peratom=True)
        logger()
        dyn.attach(logger, interval=25)
        if writetraj:
            if cpulayout is None:
                tname = "longVerlet-serial-%d-%d-%d.traj" % tuple(boundary)
            else:
                tname = ("longVerlet-%d-%d-%d--%d-%d-%d.traj" %
                         (tuple(cpulayout) + tuple(boundary)))
            traj = Trajectory(tname, "w", atoms)
            traj.write()
            dyn.attach(traj, interval=1000)
        
        for i in range(40):
            dyn.run(100)
            if i % 5 == 4:
                print(i+1, "%")
            e = (atoms.get_potential_energy() + atoms.get_kinetic_energy())/natoms
            ReportTest("Step "+str(i), e, e_start, 3e-4)





from asap3.optimize import FIRE
from asap3.md.nptberendsen import Inhomogeneous_NPTBerendsen
from ase import units


def minimize_energy(atoms, nstep, pressure_interval=10, bulkmodulus = 140e9/1e5):
    """Helper function for minimizing the energy of a nanocrystalline structure.

    It minimizes the energy while eliminating the diagonal components of the stress.
    The atomic positions are optimized using the FIRE algorithm and the stress
    is minimized using the Inhomogeneous_Berendsen algorithm.

    Parameters:

    atoms: The atoms object to be energy-minimized.

    nsteps: Number of time steps in the FIRE algorithm

    pressure_interval=10: (optional) How often to call the stress minimizer.

    bulkmodulus=1.4e6: (optional).  The bulk modulus (in bar !!!) used by the
    stress optimizer.  This value is for Cu and is useful for all metals.
    The value is uncritical, but the order of magnitude should be reasonable. 
    """
    dyn = FIRE(atoms)
    unstress = Inhomogeneous_NPTBerendsen(atoms, 50*units.fs, 0,
                                          taup=5000*units.fs,
                                          compressibility=1/bulkmodulus)
    dyn.attach(unstress.scale_positions_and_cell, interval=10)
    dyn.run(steps=nstep)





from ase.optimize.optimize import Optimizer as ase_Optimizer
from asap3 import parallelpossible as asap_parallel
if asap_parallel:
    from asap3.mpi import world


class asap_Optimizer_mixin:
    def converged(self, forces=None):
        """Did the optimization converge?"""
        if forces is None:
            forces = self.atoms.get_forces()
        else:
            forces = forces.reshape(-1, 3)
        mxf2 = (forces**2).sum(axis=1).max()
        if asap_parallel:
            mxf2 = world.max(mxf2)
        return mxf2 < self.fmax**2

class Optimizer(asap_Optimizer_mixin, ase_Optimizer):
    pass



'Test that the memory_usage() function does not crash.'

from asap3 import EMT, memory_usage
from ase.lattice.cubic import FaceCenteredCubic
from asap3.test.pytest_markers import serial

import pytest

@serial
def test_memory_usage():
    print("Making atoms")
    atoms = FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]],
                            size=(25,25,25), symbol="Cu")
    memory_usage(atoms)

    print("Attaching EMT potential")
    atoms.calc = EMT()
    memory_usage(atoms)

    print("Calculating forces")
    atoms.get_forces()
    memory_usage(atoms)

    print("Test passed!")



from asap3 import EMT, FullNeighborList
from ase.lattice.cubic import FaceCenteredCubic
import numpy as np
import ase.data
from asap3.test.pytest_markers import serial

import pytest


element = "Cu"
latconst = ase.data.reference_states[ase.data.atomic_numbers[element]]['a']
nndist = latconst / np.sqrt(2)

sizes = [(3,3,3), (2, 2, 2), (1, 1, 1)]
withpotential = (False, True)

@pytest.mark.core
@serial
@pytest.mark.parametrize('size', sizes)
@pytest.mark.parametrize('usepot', withpotential)
def test_neighborlist_small(size, usepot):
        shells = (1, 2, 3, 4)
        expected = (12, 18, 42, 54)

        print(f'Size: {size}  -  use potential: {usepot}')
        atoms = FaceCenteredCubic(
            directions=[[1,0,0],[0,1,1],[0,0,1]], 
            size=size,
            symbol=element
        )
        
        if usepot:
            atoms.calc = EMT()
            epot = atoms.get_potential_energy()

        for shell, expect in zip(shells, expected):
            cutoff = nndist * (np.sqrt(shell) + np.sqrt(shell + 1)) / 2
            nbl = FullNeighborList(cutoff, atoms, 0.0)
            allreal = True
            assert len(nbl) == len(atoms)
            # Test short neighbor lists
            for i, lst in enumerate(nbl):
                # print(f'Atom {i} has {len(lst)} neighbors: {str(lst)}')
                assert len(lst) == expect
                allreal *= max(lst) < len(atoms)
            print(f'Test passed for shell number {shell}  (all neighbors real: {allreal})')
            assert allreal

            # Test full interface
            allreal = True
            for i in range(len(atoms)):
                indices, vectors, dist2 = nbl.get_neighbors(i)
                allreal *= max(indices) < len(atoms)
                assert vectors.shape == (expect, 3)
                sumvec = vectors.sum(axis=0)
                assert sumvec.shape == (3,)
                assert np.isclose(sumvec, 0, atol=1e-8).all()
                # Convert dist2 to shell number
                shnum = dist2 / nndist**2
                shnum_int = np.round(shnum)
                assert np.isclose(shnum, shnum_int, atol=1e-6).all()
                assert shnum_int.max() == shell
            print(f'Test 2 passed for shell number {shell}  (all neighbors real: {allreal})')
            assert allreal

        if size == (1, 1, 1) and usepot:
            # Bonus test: We have an object with and EMT
            # potential attached, and minimum image convention disabled.

            nblist = atoms.calc.get_neighborlist()
            print("Length of EMT neighbor list:", len(nblist))
            for i, nbl in enumerate(nblist):
                print(f'Atom {i} has {len(nbl)} neighbors.')
                assert max(nbl) < len(atoms)



"""A subject base class"""

class Subject:
    """A subject base class"""
    def __init__(self):
        self.observers = []
        self.nsteps = 0

    def attach(self, function, interval=1, *args, **kwargs):
        """Attach callback function.

        At every *interval* steps, call *function* with arguments
        *args* and keyword arguments *kwargs*."""

        if not callable(function):
            function = function.write
        self.observers.append((function, interval, args, kwargs))

    def call_observers(self):
        self.nsteps += 1
        for function, interval, args, kwargs in self.observers:
            if self.nsteps % interval == 0:
                function(*args, **kwargs)




"""OptimizationDatabase - target values for parameter optimization.

This file contains a database of expermental and DFT-derived values
useful for optimizing classical potentials.

There are two dictionaries, elements and alloys, defining data for 
pure elements and for alloys, respectively.  The keys are chemical 
symbols.  The values are dictionaries of data.

The value dictionaries:  The keys are standard property names,
defined below.  If it makes sense, the string '_0' or '_300' is 
appended to the name to indicate the approximate temperature.
The values are tuples of (value, type, source).

The values should be given in standard units.  The type is either 
EXP (experimental) or DFT-PBE (calculated with DFT-PBE).  The 
source is a string (define a variable, so they can be reused!) 
giving the reference or a descriptive text.

Lattice constants are NOT included in the database, they are available
in ASE.

Data is accessed throught the access function XXXXX.

Property names and units::

B
    Bulk modulus (GPa)

C11, C12, C44
    Cubic elastic constants (GPa)
    
C11, C12, C13, C33, C44
    Hexagonal elastic constants (GPa)
    
Ecoh
    Cohesive energy (eV)  
    
Evac
    Vacancy formation energy (eV)
    
E100, E111
    Surface energies of 100, 111 (etc) surfaces (J/m^2)
    
Eisf
    Intrinsic stacking fault energy (J/m^2)
    
Ehcpfcc
    Energy difference between HCP and FCC structure, both with
    their optimal lattice constants
    
Ebccfcc
    Similar, but BCC minus FCC energy.
    
"""

import ase.data

##################################################
###                                            ###
### List of variables defining the references  ###
###                                            ###
##################################################

Frederikse = 'H.P.R Frederikse: "Elastic constants of single crystals", in CRC Handbook of Chemistry and Physics, 91st edition (2010).'
webelements = "www.webelements.com"
Kittel = 'C. Kittel: "Introduction to Solid State Physics", 8th ed.'
Triftshauser = "TRIFTSHAUSER, W., & MCGERVEY, J. (1975). Monovacancy Formation Energy in Copper, Silver, and Gold by Positron-Annihilation. Applied Physics, 6(2), 177�180."
JS_Surf = "Calculated by J. Schi�tz with GPAW (PBE), see /home/camp/schiotz/simulations/PotentialFit/SurfaceEnergies"

#####################################
###                               ###
### Properties for pure elements  ###
###                               ###
#####################################

elements = \
{
    'Cu': {'B_300': (140, 'EXP', webelements),
           'C11_300': (168.3, 'EXP', Frederikse),
           'C12_300': (122.1, 'EXP', Frederikse),
           'C44_300': (75.7, 'EXP', Frederikse),
           'C11_0': (176.2, 'EXP', Kittel),
           'C12_0': (124.9, 'EXP', Kittel),
           'C44_0': (81.8, 'EXP', Kittel),
           'B_0' : True,  # Can be calculated
           'Ecoh': (3.49, 'EXP', Kittel),
           'Evac': (1.29, 'EXP', Triftshauser),
           'E100': (2.888, 'PBE', JS_Surf),
           'E111': (2.635, 'PBE', JS_Surf),
           },
    'Ag': {
           'Evac': (1.29, 'EXP', Triftshauser),
           },
    'Au': {
           'Evac': (0.97, 'EXP', Triftshauser),
           },
}


#########################
###                   ###
###  Access function  ###
###                   ###
#########################

def get_data(element, property, T=0):
    """Get a data item from the database."""
    elemdata = elements[element]
    data = None
    for key in [property, "%s_%i" % (property, T), property+'_0', property+'_300']:
        if key in elemdata:
            data = elemdata[key]
            break
    if data is True:
        return calculate_data(element, key)
    elif data is None:
        # Not found.
        if property in ['a', 'c/a']:
            z = ase.data.atomic_numbers[element]
            return ase.data.reference_states[z][property]
        else:
            raise KeyError("Property %s for %s not found in database." 
                           % (property, element))
    else:
        return data[0]
    
def calculate_data(element, key):
    """Calculate data derived from other data points."""
    if '_' in key:
        key, suffix = key.split('_')
        suffix = '_' + suffix
    else:
        suffix = ""
    if key == 'B':
        return calculate_B(element, suffix)
    else:
        raise RuntimeError("Cannot calculate quantity " + key + suffix)
    
def calculate_B(element, suffix):
    z = ase.data.atomic_numbers[element]
    struct = ase.data.reference_states[z]['symmetry']
    elemdata = elements[element]
    if struct == 'fcc':
        C11 = elemdata['C11'+suffix][0]
        C12 = elemdata['C12'+suffix][0]
        return (C11 + 2*C12) / 3.0
    else:
        raise RuntimeError(f"Cannot calculate B for {element} in structure {struct}")



from ase.md.nvtberendsen import NVTBerendsen as _NVTBerendsen
from asap3.md.md import ParallelMolDynMixinNoData

class NVTBerendsen(_NVTBerendsen, ParallelMolDynMixinNoData):
    def __init__(self, atoms, timestep, temperature=None, taut=None, *,
                 fixcm=False, **kwargs):
        """Berendsen (constant N, V, T) molecular dynamics.

        Parameters:

        atoms: Atoms object
            The list of atoms.

        timestep: float
            The time step in ASE time units.

        temperature: float
            The desired temperature, in Kelvin.

        temperature_K: float
            Alias for *temperature*

        taut: float
            Time constant for Berendsen temperature coupling in ASE
            time units.

        fixcm: bool (optional)
            If True, the position and momentum of the center of mass is
            kept unperturbed.  Default: False.  This is not supported in
            parallel simulations, and should in general be unnecessary unless
            a global non-zero center-of-mass momentum needs to be preserved,
            as a zero center-of-mass momentum will remain zero.

        trajectory: Trajectory object or str (optional)
            Attach trajectory object.  If *trajectory* is a string a
            Trajectory will be constructed.  Use *None* for no
            trajectory.

        logfile: file object or str (optional)
            If *logfile* is a string, a file with that name will be opened.
            Use '-' for stdout.

        loginterval: int (optional)
            Only write a log line for every *loginterval* time steps.
            Default: 1

        append_trajectory: boolean (optional)
            Defaults to False, which causes the trajectory file to be
            overwriten each time the dynamics is restarted from scratch.
            If True, the new structures are appended to the trajectory
            file instead.

        """
        if (fixcm and getattr(atoms, 'parallel', False)):
            raise NotImplementedError("NVTBerendsen does not support fixcm=True for parallel simulations.")
        super().__init__(atoms, timestep, temperature, taut, fixcm, **kwargs)

    def run(self, steps):
        self.before_run()
        super().run(steps)
        self.after_run()




"Main neighborlist test - tests consistency between half and full lists."
from asap3 import Atoms, EMT, NeighborList, FullNeighborList
from ase.lattice.cubic import FaceCenteredCubic
import ase.data
from asap3.test.pytest_markers import ReportTest, serial
from numpy import sqrt

import pytest


def TestLists(nblist, fnb, name, natoms, count=None):
    "Run the tests on a half and a full neighbor list."
    print("")
    if count:
        print(f"Testing {name}: Length of lists")
        sum = 0
        for lst in nblist:
            sum += len(lst)
        ReportTest("   Half list", sum, count * natoms, 0)

        lfnb = list(map(len, fnb))
        assert len(lfnb) == natoms
        ReportTest("   Shortest full list", min(lfnb), 2*count, 0)
        ReportTest("   Longest full list", max(lfnb), 2*count, 0)

    print ("Testing %s: Symmetry of full list; full list atoms on half-lists."
           % (name,))
    for i, nb in enumerate(fnb):
        for jj in nb:
            j = int(jj)
            assert i in fnb[j], f'Atom {j} on list {i}'
            assert (i in nblist[j]) != (j in nblist[i]), \
                f'Exactly one of atoms {j} and {i} on half-lists'

    print(f"Testing {name}: Half-list atoms on full list.")
    for i, nb in enumerate(nblist):
        for jj in nb:
            j = int(jj)
            assert j in fnb[i], f'Atom {j} on list {i} (forward)'
            assert i in fnb[j], f'Atom {i} on list {j} (reverse)'


@pytest.mark.core
@serial
def test_neighborlists():
    element = "Cu"
    latconst = ase.data.reference_states[ase.data.atomic_numbers[element]]['a']

    atoms = FaceCenteredCubic(directions=[[1,0,0],[0,1,1],[0,0,1]], size=(9,7,5),
                            symbol=element, debug=0)
    atoms.calc = EMT(minimum_image=True)
    epot = atoms.get_potential_energy()

    nblist = atoms.calc.get_neighborlist()
    count = {}
    for lst in nblist:
        n = len(lst)
        try:
            count[n] += 1
        except KeyError:
            count[n] = 1
    # print "Histogram:"
    numbers = sorted(count.keys())
    sum = 0
    for i in numbers:
        #print i, count[i]
        sum += i*count[i]

    ReportTest("Number of neighbors (EMT's NB list)", sum, 21*len(atoms), 0)

    nblist = NeighborList(latconst * 0.5 * (1/sqrt(2) + 1), atoms, 0.0)
    #nblist = NeighborCellLocator(latconst * 0.5 * (1/sqrt(2) + 1), atoms, 0.0)
    fnb = FullNeighborList(latconst * 0.5 * (1/sqrt(2) + 1), Atoms(atoms))
    TestLists(nblist, fnb, "nearest-neigbor lists (periodic)", len(atoms), 6)

    ReportTest("Energy unperturbed 1", atoms.get_potential_energy(), epot, 1e-11)
    atoms.set_positions(atoms.get_positions())
    ReportTest("Energy unperturbed 2", atoms.get_potential_energy(), epot, 1e-11)

    nblist = NeighborList(4.98409, atoms, 0.0)
    fnb = FullNeighborList(4.98409, Atoms(atoms))
    TestLists(nblist, fnb, "long neigbor lists (periodic)", len(atoms), 21)

    ReportTest("Energy unperturbed 3", atoms.get_potential_energy(), epot, 1e-11)
    atoms.set_positions(atoms.get_positions())
    ReportTest("Energy unperturbed 4", atoms.get_potential_energy(), epot, 1e-11)

    atoms = Atoms(atoms, pbc=(0,0,0))

    nblist = NeighborList(latconst * 0.5 * (1/sqrt(2) + 1), atoms, 0.0)
    fnb = FullNeighborList(latconst * 0.5 * (1/sqrt(2) + 1), Atoms(atoms))
    TestLists(nblist, fnb, "nearest-neigbor lists (non-periodic)", len(atoms))

    atoms = Atoms(atoms, pbc=(0,1,0))

    nblist = NeighborList(4.98409, atoms, 0.0)
    fnb = FullNeighborList(4.98409, Atoms(atoms))
    TestLists(nblist, fnb, "long neigbor lists (semi-periodic)", len(atoms))
